import click
import pytest

import rhg_compute_tools.cli


@pytest.fixture
def replicate_directory_structure_on_gcs_stub(mocker):
    mocker.patch.object(
        rhg_compute_tools.cli,
        "replicate_directory_structure_on_gcs",
        new=lambda *args: click.echo(";".join(args)),
    )
    mocker.patch.object(
        rhg_compute_tools.cli, "authenticated_client", new=lambda x: str(x)
    )


@pytest.fixture
def tempfl_path(tmpdir):
    """Creates a temporary file, returning its path"""
    file = tmpdir.join("file.json")
    file.write("foobar")
    return str(file)



import os

import pytest

import rhg_compute_tools

styles = [
    s.split(".")[0]
    for s in os.listdir(
        os.path.join(os.path.dirname(rhg_compute_tools.__file__), "design", "styles")
    )
]


def test_imports():
    import matplotlib.cm
    import matplotlib.colors

    import rhg_compute_tools.design

    assert isinstance(
        matplotlib.cm.get_cmap("rhg_Blues"), matplotlib.colors.LinearSegmentedColormap
    )


@pytest.mark.parametrize("style", styles)
def test_mplstyles(style):
    import matplotlib.style

    with matplotlib.style.context(style):
        pass



import functools

import dask.array
import numpy as np
import xarray as xr
from dask import distributed as dd


def dataarrays_from_delayed(futures, client=None):
    """
    Returns a list of xarray dataarrays from a list of futures of dataarrays

    Parameters
    ----------
    futures : list
        list of :py:class:`dask.delayed.Future` objects holding
        :py:class:`xarray.DataArray` objects.
    client : object, optional
        :py:class:`dask.distributed.Client` to use in gathering
        metadata on futures. If not provided, client is inferred
        from context.

    Returns
    -------
    arrays : list
        list of :py:class:`xarray.DataArray` objects with
        :py:class:`dask.array.Array` backends.

    Examples
    --------

    Given a mapped xarray DataArray, pull the metadata into memory while
    leaving the data on the workers:

    .. code-block:: python

        >>> import numpy as np

        >>> def build_arr(multiplier):
        ...     return multiplier * xr.DataArray(
        ...         np.arange(2), dims=['x'], coords=[['a', 'b']])
        ...

        >>> client = dd.Client()
        >>> fut = client.map(build_arr, range(3))
        >>> arrs = dataarrays_from_delayed(fut)
        >>> arrs[-1]  # doctest: +ELLIPSIS
        <xarray.DataArray ...(x: 2)>
        dask.array<...shape=(2,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>
        Coordinates:
          * x        (x) <U1 'a' 'b'

    This list of arrays can now be manipulated using normal xarray tools:

    .. code-block:: python

        >>> xr.concat(arrs, dim='simulation') # doctest: +ELLIPSIS
        <xarray.DataArray ...(simulation: 3, x: 2)>
        dask.array<...shape=(3, 2), dtype=int64, chunksize=(1, 2), chunktype=numpy.ndarray>
        Coordinates:
          * x        (x) <U1 'a' 'b'
        Dimensions without coordinates: simulation

        >>> client.close()
    """

    if client is None:
        client = dd.get_client()

    delayed_arrays = client.map(lambda x: x.data, futures)

    dask_array_metadata = client.gather(
        client.map(lambda x: (x.data.shape, x.data.dtype), futures)
    )

    dask_arrays = [
        dask.array.from_delayed(delayed_arrays[i], *dask_array_metadata[i])
        for i in range(len(futures))
    ]

    # using dict(x.coords) b/c gathering coords can blow up memory for some reason
    array_metadata = client.gather(
        client.map(
            lambda x: {
                "dims": x.dims,
                "coords": dict(x.coords),
                "attrs": x.attrs,
                "name": x.name,
            },
            futures,
        )
    )

    data_arrays = [
        xr.DataArray(dask_arrays[i], **array_metadata[i]) for i in range(len(futures))
    ]

    return data_arrays


def dataarray_from_delayed(futures, dim=None, client=None):
    """
    Returns a DataArray from a list of futures

    Parameters
    ----------
    futures : list
        list of :py:class:`dask.delayed.Future` objects holding
        :py:class:`xarray.DataArray` objects.
    dim : str, optional
        dimension along which to concat :py:class:`xarray.DataArray`.
        Inferred by default.
    client : object, optional
        :py:class:`dask.distributed.Client` to use in gathering
        metadata on futures. If not provided, client is inferred
        from context.

    Returns
    -------
    array : object
        :py:class:`xarray.DataArray` concatenated along ``dim`` with
        a :py:class:`dask.array.Array` backend.

    Examples
    --------

    Given a mapped xarray DataArray, pull the metadata into memory while
    leaving the data on the workers:

    .. code-block:: python

        >>> import numpy as np, pandas as pd

        >>> def build_arr(multiplier):
        ...     return multiplier * xr.DataArray(
        ...         np.arange(2), dims=['x'], coords=[['a', 'b']])
        ...

        >>> client = dd.Client()
        >>> fut = client.map(build_arr, range(3))
        >>> da = dataarray_from_delayed(
        ...     fut,
        ...     dim=pd.Index(range(3), name='simulation'))
        ...

        >>> da  # doctest: +ELLIPSIS
        <xarray.DataArray ...(simulation: 3, x: 2)>
        dask.array<...shape=(3, 2), dtype=int64, chunksize=(1, 2), chunktype=numpy.ndarray>
        Coordinates:
          * x           (x) <U1 'a' 'b'
          * simulation  (simulation) int64 0 1 2

        >>> client.close()
    """

    data_arrays = dataarrays_from_delayed(futures, client=client)
    da = xr.concat(data_arrays, dim=dim)

    return da


def datasets_from_delayed(futures, client=None):
    """
    Returns a list of xarray datasets from a list of futures of datasets

    Parameters
    ----------
    futures : list
        list of :py:class:`dask.delayed.Future` objects holding
        :py:class:`xarray.Dataset` objects.
    client : object, optional
        :py:class:`dask.distributed.Client` to use in gathering
        metadata on futures. If not provided, client is inferred
        from context.

    Returns
    -------
    datasets : list
        list of :py:class:`xarray.Dataset` objects with
        :py:class:`dask.array.Array` backends for each variable.

    Examples
    --------

    Given a mapped :py:class:`xarray.Dataset`, pull the metadata into memory
    while leaving the data on the workers:

    .. code-block:: python

        >>> import numpy as np

        >>> def build_ds(multiplier):
        ...     return multiplier * xr.Dataset({
        ...         'var1': xr.DataArray(
        ...             np.arange(2), dims=['x'], coords=[['a', 'b']])})
        ...

        >>> client = dd.Client()
        >>> fut = client.map(build_ds, range(3))
        >>> arrs = datasets_from_delayed(fut)
        >>> arrs[-1]  # doctest: +ELLIPSIS
        <xarray.Dataset>
        Dimensions:  (x: 2)
        Coordinates:
          * x        (x) <U1 'a' 'b'
        Data variables:
            var1     (x) int64 dask.array<chunksize=(2,), meta=np.ndarray>

    This list of arrays can now be manipulated using normal xarray tools:

    .. code-block:: python

        >>> xr.concat(arrs, dim='y') # doctest: +ELLIPSIS
        <xarray.Dataset>
        Dimensions:  (x: 2, y: 3)
        Coordinates:
          * x        (x) <U1 'a' 'b'
        Dimensions without coordinates: y
        Data variables:
            var1     (y, x) int64 dask.array<chunksize=(1, 2), meta=np.ndarray>

        >>> client.close()
    """

    if client is None:
        client = dd.get_client()

    data_var_keys = client.gather(
        client.map(lambda x: list(x.data_vars.keys()), futures)
    )

    delayed_arrays = [
        {k: (client.submit(lambda x: x[k].data, futures[i])) for k in data_var_keys[i]}
        for i in range(len(futures))
    ]

    dask_array_metadata = [
        {
            k: (
                client.submit(
                    lambda x: (x[k].data.shape, x[k].data.dtype), futures[i]
                ).result()
            )
            for k in data_var_keys[i]
        }
        for i in range(len(futures))
    ]

    dask_data_arrays = [
        {
            k: (
                dask.array.from_delayed(
                    delayed_arrays[i][k], *dask_array_metadata[i][k]
                )
            )
            for k in data_var_keys[i]
        }
        for i in range(len(futures))
    ]

    # using dict(x.coords) b/c gathering coords can blow up memory for some reason
    array_metadata = [
        {
            k: client.submit(
                lambda x: {
                    "dims": x[k].dims,
                    "coords": dict(x[k].coords),
                    "attrs": x[k].attrs,
                },
                futures[i],
            ).result()
            for k in data_var_keys[i]
        }
        for i in range(len(futures))
    ]

    data_arrays = [
        {
            k: (xr.DataArray(dask_data_arrays[i][k], **array_metadata[i][k]))
            for k in data_var_keys[i]
        }
        for i in range(len(futures))
    ]

    datasets = [xr.Dataset(arr) for arr in data_arrays]

    dataset_metadata = client.gather(client.map(lambda x: x.attrs, futures))

    for i in range(len(futures)):
        datasets[i].attrs.update(dataset_metadata[i])

    return datasets


def dataset_from_delayed(futures, dim=None, client=None):
    """
    Returns an :py:class:`xarray.Dataset` from a list of futures

    Parameters
    ----------
    futures : list
        list of :py:class:`dask.delayed.Future` objects holding
        :py:class:`xarray.Dataset` objects.
    dim : str, optional
        dimension along which to concat :py:class:`xarray.Dataset`.
        Inferred by default.
    client : object, optional
        :py:class:`dask.distributed.Client` to use in gathering
        metadata on futures. If not provided, client is inferred
        from context.

    Returns
    -------
    dataset : object
        :py:class:`xarray.Dataset` concatenated along ``dim`` with
        :py:class:`dask.array.Array` backends for each variable.

    Examples
    --------

    Given a mapped :py:class:`xarray.Dataset`, pull the metadata into memory
    while leaving the data on the workers:

    .. code-block:: python

        >>> import numpy as np, pandas as pd

        >>> def build_ds(multiplier):
        ...     return multiplier * xr.Dataset({
        ...         'var1': xr.DataArray(
        ...             np.arange(2), dims=['x'], coords=[['a', 'b']])})
        ...

        >>> client = dd.Client()
        >>> fut = client.map(build_ds, range(3))
        >>> ds = dataset_from_delayed(fut, dim=pd.Index(range(3), name='y'))
        >>> ds
        <xarray.Dataset>
        Dimensions:  (x: 2, y: 3)
        Coordinates:
          * x        (x) <U1 'a' 'b'
          * y        (y) int64 0 1 2
        Data variables:
            var1     (y, x) int64 dask.array<chunksize=(1, 2), meta=np.ndarray>

        >>> client.close()
    """

    datasets = datasets_from_delayed(futures, client=client)
    ds = xr.concat(datasets, dim=dim)

    return ds


def choose_along_axis(arr, axis=-1, replace=True, nchoices=1, p=None):
    """
    Wrapper on np.random.choice, but along a single dimension within a larger array

    Parameters
    ----------
    arr : np.array
        Array with more than one dimension. Choices will be drawn from along the
        ``axis`` dimension.
    axis : integer, optional
        Dimension along which to draw samples
    replace : bool, optional
        Whether to sample with replacement. Passed to :py:func:`np.random.choice`.
        Default 1.
    nchoices : int, optional
        Number of samples to draw. Must be less than or equal to the number of
        valid options if replace is False. Default 1.
    p : np.array
        Array with the same shape as ``arr`` with weights for each choice. Each
        dimension is sampled independently, so weights will be normalized to 1
        along the ``axis`` dimension.

    Returns
    -------
    sampled : np.array
        Array with the same shape as ``arr`` but with length ``nchoices`` along axis
        ``axis`` and with values chosen from the values of ``arr`` along dimension
        ``axis`` with weights ``p``.

    Examples
    --------

    Let's say we have an array with NaNs in it:

    .. code-block:: python

        >>> arr = np.arange(40).reshape(4, 2, 5).astype(float)
        >>> for i in range(4):
        ...     arr[i, :, i+1:] = np.nan
        >>> arr  # doctest: +NORMALIZE_WHITESPACE
        array([[[ 0., nan, nan, nan, nan],
                [ 5., nan, nan, nan, nan]],
               [[10., 11., nan, nan, nan],
                [15., 16., nan, nan, nan]],
               [[20., 21., 22., nan, nan],
                [25., 26., 27., nan, nan]],
               [[30., 31., 32., 33., nan],
                [35., 36., 37., 38., nan]]])


    We can set weights such that we only select from non-nan values

    .. code-block:: python

        >>> p = (~np.isnan(arr))
        >>> p = p / p.sum(axis=2).reshape(4, 2, 1)

    Now, sampling from this along the second dimension will draw from
    these values:

    .. code-block:: python

        >>> np.random.seed(1)
        >>> choose_along_axis(arr, 2, p=p, nchoices=10)  # doctest: +NORMALIZE_WHITESPACE
        array([[[ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.],
                [ 5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.,  5.]],
               [[11., 11., 10., 11., 11., 11., 10., 10., 10., 11.],
                [15., 15., 16., 16., 16., 15., 16., 16., 15., 16.]],
               [[22., 22., 20., 22., 20., 21., 22., 20., 20., 20.],
                [25., 27., 25., 25., 26., 25., 26., 25., 26., 27.]],
               [[30., 31., 32., 31., 30., 32., 32., 32., 33., 32.],
                [38., 35., 35., 38., 36., 35., 38., 36., 38., 37.]]])

    See Also
    --------
    :py:func:`np.random.choice` : 1-d version of this function
    """
    if p is None:
        p = np.ones_like(arr).astype(float) / arr.shape[axis]

    axis = axis % len(arr.shape)
    new_shape = tuple(list(arr.shape[:axis]) + [nchoices] + list(arr.shape[axis + 1 :]))
    result = np.ndarray(shape=new_shape, dtype=arr.dtype)

    for ind in np.ndindex(tuple([l for i, l in enumerate(arr.shape) if i != axis])):
        indexer = tuple(list(ind[:axis]) + [slice(None)] + list(ind[axis:]))
        result[indexer] = np.random.choice(
            arr[indexer],
            size=nchoices,
            replace=replace,
            p=p[indexer],
        )

    return result


def choose_along_dim(da, dim, samples=1, expand=None, new_dim_name=None):
    """
    Sample values from a DataArray along a dimension

    Wraps :py:func:`np.random.choice` to sample a different random index
    (or set of indices) from along dimension ``dim`` for each combination of
    elements along the other dimensions. This is very different from block
    resampling - to block resample along a dimension simply choose a set
    of indices and draw these from the array using :py:meth:`xr.DataArray.sel`.

    Parameters
    ----------
    da : xr.DataArray
        DataArray from which to sample values.
    dim: str
        Dimension along which to sample. Sampling will draw from elements along
        this dimension for all combinations of other dimensions.
    samples : int, optional
        Number of samples to take from the dimension ``dim``. If greater than 1,
        ``expand`` is ignored (and set to True).
    expand : bool, optional
        Whether to expand the array along the sampled dimension.
    new_dim_name : str, optoinal
        Name for the new dimension. If not provided, will use ``dim``.

    Returns
    -------
    sampled : xr.DataArray
        DataArray with sampled values chosen along dimension ``dim``

    Examples
    --------

    .. code-block:: python

        >>> da = xr.DataArray(
        ...     np.arange(40).reshape(4, 2, 5),
        ...     dims=['x', 'y', 'z'],
        ...     coords=[np.arange(4), np.arange(2), np.arange(5)],
        ... )

        >>> da  # doctest: +NORMALIZE_WHITESPACE
        <xarray.DataArray (x: 4, y: 2, z: 5)>
        array([[[ 0,  1,  2,  3,  4],
                [ 5,  6,  7,  8,  9]],
               [[10, 11, 12, 13, 14],
                [15, 16, 17, 18, 19]],
               [[20, 21, 22, 23, 24],
                [25, 26, 27, 28, 29]],
               [[30, 31, 32, 33, 34],
                [35, 36, 37, 38, 39]]])
        Coordinates:
          * x        (x) int64 0 1 2 3
          * y        (y) int64 0 1
          * z        (z) int64 0 1 2 3 4

    We can take a random value along the ``'z'`` dimension:

    .. code-block:: python

        >>> np.random.seed(1)
        >>> choose_along_dim(da, 'z')
        <xarray.DataArray (x: 4, y: 2)>
        array([[ 2,  8],
               [10, 16],
               [20, 25],
               [30, 36]])
        Coordinates:
          * x        (x) int64 0 1 2 3
          * y        (y) int64 0 1


    If you provide a ``sample`` argument greater than one (or
    set expand=True) the array will be expanded to a new
    dimension:

    .. code-block:: python

        >>> np.random.seed(1)
        >>> choose_along_dim(da, 'z', samples=3)  # doctest: +NORMALIZE_WHITESPACE
        <xarray.DataArray (x: 4, y: 2, z: 3)>
        array([[[ 2,  3,  0],
                [ 6,  5,  5]],
               [[10, 11, 11],
                [17, 17, 18]],
               [[21, 24, 20],
                [28, 27, 27]],
               [[30, 30, 34],
                [39, 36, 38]]])
        Coordinates:
          * x        (x) int64 0 1 2 3
          * y        (y) int64 0 1
          * z        (z) int64 0 1 2
    """
    sampled = choose_along_axis(da.values, axis=da.get_axis_num(dim), nchoices=samples)

    if samples > 1:
        expand = True

    if not expand:
        sampled = np.take(sampled, 0, axis=da.get_axis_num(dim))
        return xr.DataArray(
            sampled,
            dims=[d for d in da.dims if d != dim],
            coords=[da.coords[d] for d in da.dims if d != dim],
        )

    else:
        if new_dim_name is None:
            new_dim_name = dim

        return xr.DataArray(
            sampled,
            dims=[d if d != dim else new_dim_name for d in da.dims],
            coords=[da.coords[d] if d != dim else np.arange(samples) for d in da.dims],
        )


@xr.register_dataarray_accessor("random")
class random:
    def __init__(self, xarray_obj):
        self._xarray_obj = xarray_obj

    @functools.wraps(choose_along_dim)
    def choice(self, *args, **kwargs):
        return choose_along_dim(self._xarray_obj, *args, **kwargs)



# -*- coding: utf-8 -*-

"""Top-level package for RHG Compute Tools."""


import pkg_resources

__author__ = """Michael Delgado"""
__email__ = "mdelgado@rhg.com"
__version__ = pkg_resources.get_distribution("rhg_compute_tools").version



from __future__ import absolute_import

from rhg_compute_tools.design.colors import _load_colors
from rhg_compute_tools.design.plotting import add_colorbar, get_color_scheme

_load_colors()

# import matplotlib.font_manager as fm
#
# RHG_FONTS = [
#
#     # rhodium v1
#     # ('Univers LT', 'ttf'),
#     # ('Knockout', 'ttf'),
#
#     # rhodium 2019
#     ('Fira Sans', 'ttf'),
#     # ('DIN Pro', 'ttf'),
#     # ('CamingoDos', 'ttf'),
#     #
#     # # impactlab
#     # ('Chaparral Pro', 'ttf'),
#     # ('Merriweather', 'ttf'),
#     ]
#
# for (sysfont, ext) in RHG_FONTS:
#     fm.findfont(sysfont, fontext=ext, rebuild_if_missing=False)

from matplotlib.font_manager import _rebuild

_rebuild()

__all__ = ["get_color_scheme", "add_colorbar"]



import click

from rhg_compute_tools.gcs import (
    authenticated_client,
    replicate_directory_structure_on_gcs,
)


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def rctools_cli():
    """Rhodium Compute Tools"""
    pass


@rctools_cli.group()
def gcs():
    """Tools for interacting with Google Cloud Storage"""
    pass


@gcs.command()
@click.argument("src", type=click.Path(exists=True))
@click.argument("dst", type=click.Path())
@click.option(
    "-c",
    "--credentials",
    type=click.Path(exists=True),
    help="Optional path to GCS credentials file.",
)
def repdirstruc(src, dst, credentials):
    """Replicate directory structure onto GCS bucket.

    SRC is path to a local directory. Directories within will be replicated.
    DST is gs://[bucket] and optional path to replicate SRC into.

    If --credentials or -c is not explicitly given, checks the
    GOOGLE_APPLICATION_CREDENTIALS environment variable for path to a GCS
    credentials file, or default service accounts for authentication. See
    https://googleapis.dev/python/google-api-core/latest/auth.html for more
    details.
    """
    client = authenticated_client(credentials)
    replicate_directory_structure_on_gcs(src, dst, client)



import collections.abc
import dis
import functools
import inspect
import itertools
import json
import os
import types

import numpy as np
import toolz


def expand(func):
    """
    Decorator to expand an (args, kwargs) tuple in function calls

    Intended for use with the :py:func:`collapse` function

    Parameters
    ----------
    func : function
        Function to have arguments expanded. Func can have any
        number of positional and keyword arguments.

    Returns
    -------
    wrapped : function
        Wrapped version of ``func`` which accepts a single
        ``(args, kwargs)`` tuple.

    Examples
    --------

    .. code-block:: python

        >>> @expand
        ... def my_func(a, b, exp=1):
        ...     return (a * b)**exp
        ...

        >>> my_func(((2, 3), {}))
        6

        >>> my_func(((2, 3, 2), {}))
        36

        >>> my_func((tuple([]), {'b': 4, 'exp': 2, 'a': 1}))
        16

    This function can be used in combination with the ``collapse`` helper
    function, which allows more natural parameter calls

    .. code-block:: python

        >>> my_func(collapse(2, 3, exp=2))
        36

    These can then be paired to enable many parameterized function calls:

    .. code-block:: python

        >>> func_calls = [collapse(a, a+1, exp=a) for a in range(5)]

        >>> list(map(my_func, func_calls))
        [1, 2, 36, 1728, 160000]

    """

    @functools.wraps(func)
    def inner(ak, *args, **kwargs):
        return func(*ak[0], *args, **ak[1], **kwargs)

    return inner


def collapse(*args, **kwargs):
    """
    Collapse positional and keyword arguments into an (args, kwargs) tuple

    Intended for use with the :py:func:`expand` decorator

    Parameters
    ----------
    *args
        Variable length argument list.
    **kwargs
        Arbitrary keyword arguments.

    Returns
    -------
    args : tuple
        Positional arguments tuple
    kwargs : dict
        Keyword argument dictionary
    """
    return (args, kwargs)


def collapse_product(*args, **kwargs):
    """

    Parameters
    ----------

    *args
        Variable length list of iterables
    **kwargs
        Keyword arguments, whose values must be iterables

    Returns
    -------
    iterator
        Generator with collapsed arguments

    See Also
    --------

    Function :py:func:`collapse`

    Examples
    --------

    .. code-block:: python

        >>> @expand
        ... def my_func(a, b, exp=1):
        ...     return (a * b)**exp
        ...

        >>> product_args = list(collapse_product(
        ...     [0, 1, 2],
        ...     [0.5, 2],
        ...     exp=[0, 1]))

        >>> product_args  # doctest: +NORMALIZE_WHITESPACE
        [((0, 0.5), {'exp': 0}),
         ((0, 0.5), {'exp': 1}),
         ((0, 2), {'exp': 0}),
         ((0, 2), {'exp': 1}),
         ((1, 0.5), {'exp': 0}),
         ((1, 0.5), {'exp': 1}),
         ((1, 2), {'exp': 0}),
         ((1, 2), {'exp': 1}),
         ((2, 0.5), {'exp': 0}),
         ((2, 0.5), {'exp': 1}),
         ((2, 2), {'exp': 0}),
         ((2, 2), {'exp': 1})]

        >>> list(map(my_func, product_args))
        [1.0, 0.0, 1, 0, 1.0, 0.5, 1, 2, 1.0, 1.0, 1, 4]
    """
    num_args = len(args)
    kwarg_keys = list(kwargs.keys())
    kwarg_vals = [kwargs[k] for k in kwarg_keys]

    format_iterations = lambda x: (
        tuple(x[:num_args]),
        dict(zip(kwarg_keys, x[num_args:])),
    )

    return map(format_iterations, itertools.product(*args, *kwarg_vals))


class NumpyEncoder(json.JSONEncoder):
    """
    Helper class for json.dumps to coerce numpy objects to native python
    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.int64):
            return int(obj)
        elif isinstance(obj, np.float64):
            return float(obj)
        elif isinstance(obj, np.int32):
            return int(obj)
        elif isinstance(obj, np.float32):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


def checkpoint(
    jobs,
    futures,
    job_name,
    log_dir=".",
    extra_pending=None,
    extra_errors=None,
    extra_others=None,
):
    """
    checkpoint and save a job state to disk
    """

    err_msg = "lengths do not match: jobs [{}] != futures [{}]".format(
        len(jobs), len(futures)
    )

    assert len(jobs) == len(futures), err_msg

    if extra_pending is None:
        extra_pending = []

    if extra_errors is None:
        extra_errors = []

    if extra_others is None:
        extra_others = {}

    pending_jobs = [
        jobs[i] for i, f in enumerate(futures) if f.status == "pending"
    ] + extra_pending

    errored_jobs = [
        jobs[i] for i, f in enumerate(futures) if f.status == "error"
    ] + extra_errors

    other_jobs = {}

    for i, f in enumerate(futures):
        if f.status in ["pending", "error", "finished"]:
            continue

        if f.status not in other_jobs:
            other_jobs[f.status] = []

        other_jobs[f.status].append(jobs[i])

    for k, v in extra_others.items():
        if k not in other_jobs:
            other_jobs[k] = []

        other_jobs[k].append(v)

    with open(os.path.join(log_dir, "{}.pending".format(job_name)), "w+") as f:
        f.write(json.dumps(pending_jobs, cls=NumpyEncoder))

    with open(os.path.join(log_dir, "{}.err".format(job_name)), "w+") as f:
        f.write(json.dumps(errored_jobs, cls=NumpyEncoder))

    with open(os.path.join(log_dir, "{}.other".format(job_name)), "w+") as f:
        f.write(json.dumps(other_jobs, cls=NumpyEncoder))


def recover(job_name, log_dir="."):
    """
    recover pending, errored, other jobs from a checkpoint
    """

    with open(os.path.join(log_dir, "{}.pending".format(job_name)), "r") as f:
        content = f.read()
        if len(content) == 0:
            pending = {}
        else:
            pending = json.loads(content)

    with open(os.path.join(log_dir, "{}.err".format(job_name)), "r") as f:
        content = f.read()
        if len(content) == 0:
            errored = {}
        else:
            errored = json.loads(content)

    with open(os.path.join(log_dir, "{}.other".format(job_name)), "r") as f:
        content = f.read()
        if len(content) == 0:
            other = {}
        else:
            other = json.loads(content)

    return pending, errored, other


class html(object):
    def __init__(self, body):
        self.body = body

    def _repr_html_(self):
        return self.body


_default_allowed_types = (
    types.FunctionType,
    types.ModuleType,
    (type if not hasattr(types, "ClassType") else types.ClassType),
    types.MethodType,
    types.BuiltinMethodType,
    types.BuiltinFunctionType,
)


@toolz.functoolz.curry
def block_globals(obj, allowed_types=None, include_defaults=True, whitelist=None):
    """
    Decorator to prevent globals and undefined closures in functions and classes

    Parameters
    ----------
    func : function
        Function to decorate. All globals not matching one of the allowed
        types will raise an AssertionError
    allowed_types : type or tuple of types, optional
        Types which are allowed as globals. By default, functions and
        modules are allowed. The full set of allowed types is drawn from
        the ``types`` module, and includes :py:class:`~types.FunctionType`,
        :py:class:`~types.ModuleType`, :py:class:`~types.MethodType`,
        :py:class:`~types.ClassType`,
        :py:class:`~types.BuiltinMethodType`, and
        :py:class:`~types.BuiltinFunctionType`.
    include_defaults : bool, optional
        If allowed_types is provided, setting ``include_defaults`` to True will
        append the default list of functions, modules, and methods to the
        user-passed list of allowed types. Default is True, in which case
        only the user-passed elements will be allowed. Setting to False will
        allow only the types passed in ``allowed_types``.
    whitelist : list of str, optional
        Optional list of variable names to whitelist. If a list is provided,
        global variables will be compared to elements of this list based on
        their string names. Default (None) is no whitelist.

    Examples
    --------

    Wrap a function to block globals:

    .. code-block:: python

        >>> my_data = 10

        >>> @block_globals
        ... def add_5(data):
        ...     ''' can you spot the global? '''
        ...     a_number = 5
        ...     result = a_number + my_data
        ...     return result  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Illegal <class 'int'> global found in add_5: my_data

    Wrapping a class will prevent globals from being used in all methods:

    .. code-block:: python

        >>> @block_globals
        ... class MyClass:
        ...
        ...     @staticmethod
        ...     def add_5(data):
        ...         ''' can you spot the global? '''
        ...         a_number = 5
        ...         result = a_number + my_data
        ...         return result  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Illegal <class 'int'> global found in add_5: my_data

    By default, functions and modules are allowed in the list of globals. You
    can modify this list with the ``allowed_types`` argument:

    .. code-block:: python

        >>> result_formatter = 'my number is {}'
        >>> @block_globals(allowed_types=str)
        ... def add_5(data):
        ...     ''' only allowed globals here! '''
        ...     a_number = 5
        ...     result = a_number + data
        ...     return result_formatter.format(result)
        ...
        >>> add_5(3)
        'my number is 8'

    block_globals will also catch undefined references:

    .. code-block:: python

        >>> @block_globals
        ... def get_mean(df):
        ...     return da.mean()  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Undefined global in get_mean: da
    """

    if allowed_types is None:
        allowed_types = _default_allowed_types

    if (allowed_types is not None) and include_defaults:
        if not isinstance(allowed_types, collections.abc.Sequence):
            allowed_types = [allowed_types]

        allowed_types = tuple(list(allowed_types) + list(_default_allowed_types))

    if whitelist is None:
        whitelist = []

    if isinstance(obj, type):
        for attr in obj.__dict__:
            if callable(getattr(obj, attr)):
                setattr(obj, attr, block_globals(getattr(obj, attr)))
        return obj

    closurevars = inspect.getclosurevars(obj)
    for instr in dis.get_instructions(obj):
        if instr.opname == "LOAD_GLOBAL":
            if instr.argval in closurevars.builtins:
                continue
            elif (instr.argval in closurevars.globals) or (
                instr.argval in closurevars.nonlocals
            ):
                if instr.argval in whitelist:
                    continue
                if instr.argval in closurevars.globals:
                    g = closurevars.globals[instr.argval]
                else:
                    g = closurevars.nonlocals[instr.argval]
                if not isinstance(g, allowed_types):
                    raise TypeError(
                        "Illegal {} global found in {}: {}".format(
                            type(g),
                            obj.__name__,
                            instr.argval,
                        )
                    )
            else:
                raise TypeError(
                    "Undefined global in {}: {}".format(
                        obj.__name__,
                        instr.argval,
                    )
                )

    @functools.wraps(obj)
    def inner(*args, **kwargs):
        return obj(*args, **kwargs)

    return inner



import requests
from ruamel import yaml

# TODO: change this branch to master when gateway stuff is merged
r = requests.get(
    "https://raw.githubusercontent.com/RhodiumGroup/helm-chart/gateway-tweaks/values.yml"
)

data = yaml.safe_load(r.content)["dask-gateway"]["gateway"]["extraConfig"][
    "optionHandler"
]

# change float cores to int cores b/c float not allowed for local cluster

data = "from math import ceil\nfrom dask_gateway_server.options import Integer\n" + data
data = data.replace('Float("cpus"', 'Integer("cpus"').replace(
    'Float("scheduler_cores", default=3.5', 'Integer("scheduler_cores", default=1'
)

with open("dask_gateway_config.py", "w") as f:
    f.write(data)



"""
White-box testing to ensure that CLI passes variables correctly
"""


import click
import pytest
from click.testing import CliRunner

import rhg_compute_tools.cli
from tests.fixtures import replicate_directory_structure_on_gcs_stub, tempfl_path


@pytest.mark.parametrize("credflag", ["-c", "--credentials", None])
def test_repdirstruc(
    credflag,
    replicate_directory_structure_on_gcs_stub,
    tmpdir,
    tempfl_path,
    monkeypatch,
):
    """Test rctools gcs repdirstruc for main input options"""
    # Setup CLI args
    cred_path = str(tempfl_path)
    credargs = [credflag, cred_path]
    src_path = str(tmpdir)
    dst_path = "gcs://foo/bar"

    if credflag is None:
        credargs = []
        cred_path = str(None)

    # Run CLI
    runner = CliRunner()
    result = runner.invoke(
        rhg_compute_tools.cli.rctools_cli,
        ["gcs", "repdirstruc"] + credargs + [src_path, dst_path],
    )

    expected_output = ";".join([src_path, dst_path, cred_path]) + "\n"
    assert result.output == expected_output


@pytest.mark.parametrize("credflag", ["-c", "--credentials"])
def test_repdirstruc_nocredfile(
    credflag, replicate_directory_structure_on_gcs_stub, tmpdir, monkeypatch
):
    """Test rctools gcs repdirstruc for graceful fail when cred file missing"""
    # Setup CLI args
    cred_path = "_foobar.json"
    credargs = [credflag, cred_path]
    src_path = str(tmpdir)
    dst_path = "gcs://foo/bar"

    # Run CLI
    runner = CliRunner()
    result = runner.invoke(
        rhg_compute_tools.cli.rctools_cli,
        ["gcs", "repdirstruc"] + credargs + [src_path, dst_path],
    )

    expected = (
        "Usage: rctools-cli gcs repdirstruc [OPTIONS] SRC DST\nTry"
        " 'rctools-cli gcs repdirstruc -h' for help.\n\nError: "
        "Invalid value for '-c' / '--credentials': Path "
        "'_foobar.json' does not exist.\n"
    )
    assert result.output.replace('"', "'") == expected.replace('"', "'")



# -*- coding: utf-8 -*-

"""Tools for interacting with GCS infrastructure."""

import os
import shlex
import subprocess
from datetime import datetime as dt
from os.path import basename, exists, isdir, join

from google.cloud import storage
from google.oauth2 import service_account


def authenticated_client(credentials=None, **client_kwargs):
    """Convenience function to create an authenticated GCS client.

    Parameters
    ----------
    credentials : str or None, optional
        Str path to storage credentials authentication file. If None
        is passed (default) will create a Client object with no args, using
        the authorization credentials for the current environment. See the
        [google cloud storage docs](
        https://googleapis.dev/python/google-api-core/latest/auth.html)
        for an overview of the authorization options.
    client_kwargs : optional
        kwargs to pass to the `get_client` function

    Returns
    -------
    google.cloud.storage.Client
    """
    if credentials is None:
        client = storage.Client()
    else:
        creds = service_account.Credentials.from_service_account_file(str(credentials))
        client = storage.Client(credentials=creds, **client_kwargs)

    return client


def get_bucket(
    credentials=None, bucket_name="rhg-data", return_client=False, **client_kwargs
):
    """Return a bucket object from Rhg's GCS system.

    Parameters
    ----------
    credentials : str or None, optional
        Str path to storage credentials authentication file. If None
        is passed (default) will create a Client object with no args, using
        the authorization credentials for the current environment. See the
        [google cloud storage docs](
        https://googleapis.dev/python/google-api-core/latest/auth.html)
        for an overview of the authorization options.
    bucket_name : str, optional
        Name of bucket. Typically, we work with ``rhg_data`` (default)
    return_client : bool, optional
        Return the Client object as a second object.
    client_kwargs : optional
        kwargs to pass to the `get_client` function

    Returns
    -------
    bucket : :py:class:`google.cloud.storage.bucket.Bucket`
    """
    client = authenticated_client(credentials=credentials, **client_kwargs)
    result = client.get_bucket(bucket_name)

    if return_client:
        result = (result, client)

    return result


def _remove_prefix(text, prefix="/gcs/rhg-data/"):
    return text[text.startswith(prefix) and len(prefix) :]


def _get_path_types(src, dest):
    src_gs = src
    src_gcs = src
    dest_gs = dest
    dest_gcs = dest
    if src_gs.startswith("/gcs/"):
        src_gs = src.replace("/gcs/", "gs://")
    if src_gcs.startswith("gs://"):
        src_gs = src.replace("gs://", "/gcs/")
    if dest_gs.startswith("/gcs/"):
        dest_gs = dest.replace("/gcs/", "gs://")
    if dest_gcs.startswith("gs://"):
        dest_gcs = dest.replace("gs://", "/gcs/")

    return src_gs, src_gcs, dest_gs, dest_gcs


def rm(path, flags=[]):
    """Remove a file or recursively remove a directory from local
    path to GCS or vice versa. Must have already authenticated to use.
    Notebook servers are automatically authenticated, but workers
    need to pass the path to the authentication json file to the
    GOOGLE_APPLICATION_CREDENTIALS env var.
    This is done automatically for rhg-data.json when using the get_worker
    wrapper.

    Parameters
    ----------
    path : str or :class:`pathlib.Path`
        The path to the source and destination file or directory.
        Either the `/gcs` or `gs:/` prefix will work.
    flags : list of str, optional
        String of flags to add to the gsutil rm command. e.g.
        `flags=['r']` will run the command `gsutil -m rm -r...`
        (recursive remove)

    Returns
    -------
    str
        stdout from gsutil call
    str
        stderr from gsutil call
    :py:class:`datetime.timedelta`
        Time it took to copy file(s).
    """

    st_time = dt.now()

    path = str(path).replace("/gcs/", "gs://")

    cmd = "gsutil -m rm " + " ".join(["-" + f for f in flags]) + f" {path}"

    print(f"Running cmd: {cmd}")
    cmd = shlex.split(cmd)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    end_time = dt.now()

    return stdout, stderr, end_time - st_time


def replicate_directory_structure_on_gcs(src, dst, client):
    """
    Replicate a local directory structure on google cloud storage

    Parameters
    ----------
    src : str
        Path to the root directory on the source machine. The directory
        structure within this directory will be reproduced within `dst`,
        e.g. `/Users/myusername/my/data`
    dst : str
        A url for the root directory of the destination, starting with
        `gs://[bucket_name]/`, e.g. `gs://my_bucket/path/to/my/data`
    client : google.cloud.storage.client.Client
        An authenticated :py:class:`google.cloud.storage.client.Client` object.
    """
    if dst.startswith("gs://"):
        dst = dst[5:]
    elif dst.startswith("gcs://"):
        dst = dst[6:]
    else:
        raise ValueError("dst must begin with `gs://` or `gcs://`")

    bucket_name = dst.split("/")[0]
    blob_path = "/".join(dst.split("/")[1:])

    bucket = client.get_bucket(bucket_name)

    for d, dirnames, files in os.walk(src):
        dest_path = os.path.join(blob_path, os.path.relpath(d, src))

        # make sure there is exactly one trailing slash:
        dest_path = dest_path.rstrip("/") + "/"

        # ignore "." directory
        if dest_path == "./":
            continue

        blob = bucket.blob(dest_path)
        blob.upload_from_string("")


def cp(src, dest, flags=[]):
    """Copy a file or recursively copy a directory from local
    path to GCS or vice versa. Must have already authenticated to use.
    Notebook servers are automatically authenticated, but workers
    need to pass the path to the authentication json file to the
    GOOGLE_APPLICATION_CREDENTIALS env var.
    This is done automatically for rhg-data.json when using the get_worker
    wrapper.

    Parameters
    ----------
    src, dest : str
        The paths to the source and destination file or directory.
        If on GCS, either the `/gcs` or `gs:/` prefix will work.
    flags : list of str, optional
        String of flags to add to the gsutil cp command. e.g.
        `flags=['r']` will run the command `gsutil -m cp -r...`
        (recursive copy)

    Returns
    -------
    str
        stdout from gsutil call
    str
        stderr from gsutil call
    :py:class:`datetime.timedelta`
        Time it took to copy file(s).
    """

    st_time = dt.now()

    src = str(src)
    dest = str(dest)

    # make sure we're using URL
    # if /gcs or gs:/ are not in src or not in dest
    # then these won't change anything
    src_gs, src_gcs, dest_gs, dest_gcs = _get_path_types(src, dest)

    # if directory already existed cp would put src into dest_gcs
    if exists(dest_gcs):
        dest_base = join(dest_gcs, basename(src))
    # else cp would have put the contents of src into the new directory
    else:
        dest_base = dest_gcs

    cmd = (
        "gsutil -m cp "
        + " ".join(["-" + f for f in flags])
        + " {} {}".format(src_gs, dest_gs)
    )

    print(f"Running cmd: {cmd}")
    cmd = shlex.split(cmd)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    # need to add directories if you were recursively copying a directory
    if isdir(src_gcs) and dest_gcs.startswith("/gcs/"):
        # now make directory blobs on gcs so that gcsfuse recognizes it
        dirs_to_make = [x[0].replace(src, dest_base) for x in os.walk(src)]
        for d in dirs_to_make:
            os.makedirs(d, exist_ok=True)

    end_time = dt.now()

    return stdout, stderr, end_time - st_time


def sync(src, dest, flags=["r", "d"]):
    """Sync a directory from local to GCS or vice versa. Uses `gsutil rsync`.
    Must have already authenticated to use. Notebook servers
    are automatically authenticated, but workers need to pass the path
    to the authentication json file to the GCLOUD_DEFAULT_TOKEN_FILE env var.
    This is done automatically for rhg-data.json when using the get_worker
    wrapper.

    Parameters
    ----------
    src, dest : str
        The paths to the source and destination file or directory.
        If on GCS, either the `/gcs` or `gs:/` prefix will work.
    flags : list of str, optional
        String of flags to add to the gsutil cp command. e.g.
        `flags=['r','d']` will run the command `gsutil -m cp -r -d...`
        (recursive copy, delete any files on dest that are not on src).
        This is the default set of flags.

    Returns
    -------
    str
        stdout from gsutil call
    str
        stderr from gsutil call
    :py:class:`datetime.timedelta`
        Time it took to copy file(s).
    """

    st_time = dt.now()

    src = str(src)
    dest = str(dest)

    # remove trailing /'s
    src = src.rstrip("/")
    dest = dest.rstrip("/")

    # make sure we're using URL
    # if /gcs or gs:/ are not in src or not in dest
    # then these won't change anything
    src_gs, src_gcs, dest_gs, dest_gcs = _get_path_types(src, dest)

    cmd = (
        "gsutil -m rsync "
        + " ".join(["-" + f for f in flags])
        + " {} {}".format(src_gs, dest_gs)
    )

    print(f"Running cmd: {cmd}")

    cmd = shlex.split(cmd)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    # need to add directories if you were recursively copying a directory TO
    # gcs now make directory blobs on gcs so that gcsfuse recognizes it
    if dest_gcs.startswith("/gcs/"):
        dirs_to_make = [x[0].replace(src_gcs, dest_gcs) for x in os.walk(src_gcs)]
        for d in dirs_to_make:
            os.makedirs(d, exist_ok=True)

    end_time = dt.now()

    return stdout, stderr, end_time - st_time



import os

import matplotlib.cm
import matplotlib.style
import numpy as np
from matplotlib.colors import LinearSegmentedColormap, ListedColormap

RHG_COLOR_GRID = np.array(
    [
        ["#ACD8F1", "#63AAD6", "#0078AD", "#055A7F", "#023B56"],
        ["#C1E399", "#A0D55F", "#77B530", "#59901B", "#366108"],
        ["#FDE8A5", "#FBD568", "#E7B731", "#C49C20", "#926E00"],
        ["#F9C7A2", "#FEA569", "#E97625", "#C45506", "#913F05"],
        ["#ECA5AB", "#E66967", "#C32524", "#920605", "#6C0405"],
        ["#D7BBE3", "#BE95CF", "#915FA4", "#633A76", "#6C0405"],
    ]
).T

_custom_continuous_cmaps = {
    "cil_RdBu": [
        "#e0603f",
        "#ed7453",
        "#f68f50",
        "#ffaa4d",
        "#ffcb58",
        "#ffea80",
        "#ffffad",
        "#d9ffd9",
        "#bbeae4",
        "#88d8da",
        "#55c7d2",
        "#41bad1",
        "#41a9c1",
        "#4197b0",
    ],
    "cil_Reds": [
        "#FFDB6C",
        "#ffcb58",
        "#FFBB53",
        "#ffaa4d",
        "#FB9D4F",
        "#f68f50",
        "#F28252",
        "#ed7453",
        "#E76A49",
        "#e0603f",
        "#D45433",
        "#c84726",
        "#b53919",
    ],
    "cil_Blues": [
        "#A2E1DF",
        "#88d8da",
        "#6FD0D6",
        "#55c7d2",
        "#4BC1D2",
        "#41bad1",
        "#41B2C9",
        "#41a9c1",
        "#41A0B9",
        "#4197b0",
        "#398AA1",
        "#307c92",
        "#065b74",
    ],
    "rhg_Blues": RHG_COLOR_GRID[:, 0],
    "rhg_Greens": RHG_COLOR_GRID[:, 1],
    "rhg_Yellows": RHG_COLOR_GRID[:, 2],
    "rhg_Oranges": RHG_COLOR_GRID[:, 3],
    "rhg_Reds": RHG_COLOR_GRID[:, 4],
    "rhg_Purples": RHG_COLOR_GRID[:, 5],
}

_custom_discrete_cmaps = {
    "rhg_standard": RHG_COLOR_GRID[2, :],
    "rhg_light": RHG_COLOR_GRID[1, :],
}


def _load_colors():
    for cmap_name, cmap_colors in _custom_continuous_cmaps.items():
        cmap = LinearSegmentedColormap.from_list(cmap_name, cmap_colors)
        matplotlib.cm.register_cmap(cmap=cmap)
        cmap_r = cmap.reversed()
        matplotlib.cm.register_cmap(cmap=cmap_r)

    for cmap_name, cmap_colors in _custom_discrete_cmaps.items():
        cmap = ListedColormap(cmap_colors, name=cmap_name)
        matplotlib.cm.register_cmap(cmap=cmap)
        cmap_r = cmap.reversed()
        matplotlib.cm.register_cmap(cmap=cmap_r)

    matplotlib.style.core.USER_LIBRARY_PATHS += [
        os.path.join(os.path.dirname(__file__), "styles")
    ]

    matplotlib.style.core.reload_library()



# -*- coding: utf-8 -*-
"""Tools for interacting with kubernetes."""

import os
import socket
import traceback as tb
import warnings
from collections.abc import Sequence
from pathlib import Path

import dask
import numpy as np
import yaml as yml
from dask import distributed as dd

# is dask-gateway available
GATEWAY = False
try:
    import dask_gateway

    GATEWAY = True
except ModuleNotFoundError:
    from dask_kubernetes import KubeCluster


def traceback(ftr):
    return tb.print_exception(type(ftr.exception()), ftr.exception(), ftr.traceback())


def _append_docstring():
    func_with_docstring = _get_cluster_dask_kubernetes
    if GATEWAY:
        func_with_docstring = _get_cluster_dask_gateway

    def decorator(func):
        if func.__doc__ is None:
            func.__doc__ = ""
        if func_with_docstring.__doc__ is None:
            func.__doc__ = ""

        func.__doc__ = func.__doc__ + "\n".join(
            func_with_docstring.__doc__.lstrip().split("\n")[1:]
        )

        return func

    return decorator


def _get_cluster_dask_gateway(**kwargs):
    """
    Start dask.kubernetes cluster and dask.distributed client

    All arguments are optional. If not provided, defaults will be used. To view
    defaults, instantiate a :class:`dask_gateway.Gateway` object and call
    `gateway.cluster_options()`.

    Parameters
    ----------
    name : str, optional
        Name of worker image to use (e.g. ``rhodium/worker:latest``). If ``None``
        (default), default to worker specified in ``template_path``.
    tag : str, optional
        Tag of the worker image to use. Cannot be used in combination with
        ``name``, which should include a tag. If provided, overrides the
        tag of the image specified in ``template_path``. If ``None``
        (default), the full image specified in ``name`` or ``template_path``
        is used.
    extra_pip_packages : str, optional
        Extra pip packages to install on worker. Packages are installed
        using ``pip install extra_pip_packages``.
    profile : One of ["micro", "standard", "big", "giant"]
        Determines size of worker. CPUs assigned are slightly under 1, 2, 4, and 8,
        respectively. Memory assigned is slightly over 6, 12, 24, and 48 GB,
        respectively.
    cpus : float, optional
        Set the CPUs requested for your workers as defined by ``profile``. Will
        raise error if >7.5, because our 8-CPU nodes need ~.5 vCPU for kubernetes pods.
        (NOTE 12/15/20: This is currently set to 1 by default to allow for mapping
        big workflows across inputs, see
        https://github.com/dask/dask-gateway/issues/364).
    cred_name : str, optional
        Name of Google Cloud credentials file to use, equivalent to providing
        ``cred_path='/opt/gcsfuse_tokens/{}.json'.format(cred_name)``. May not use
        if ``cred_path`` is specified.
    cred_path : str, optional
        Path to Google Cloud credentials file to use. May not use if ``cred_name`` is
        specified.
    env_items : dict, optional
        A dictionary of env variable 'name'-'value' pairs to append to the env
        variables included in ``template_path``, e.g.

        .. code-block:: python

            {
                'MY_ENV_VAR': 'some string',
            }

    extra_worker_labels : dict, optional
        Dictionary of kubernetes labels to apply to pods. None (default) results
        in no additional labels besides those in the template, as well as
        ``jupyter_user``, which is inferred from the ``JUPYTERHUB_USER``, or, if
        not set, the server's hostname.
    extra_pod_tolerations : list of dict, optional
        List of pod toleration dictionaries. For example, to match a node pool
        NoSchedule toleration, you might provide:

        .. code-block:: python

            extra_pod_tolerations=[
                {
                    "effect": "NoSchedule",
                    "key": "k8s.dask.org_dedicated",
                    "operator": "Equal",
                    "value": "worker-highcpu"
                },
                {
                    "effect": "NoSchedule",
                    "key": "k8s.dask.org/dedicated",
                    "operator": "Equal",
                    "value": "worker-highcpu"
                }
            ]

    keep_default_tolerations : bool, optional
        Whether to append (default) or replace the default tolerations. Ignored if
        ``extra_pod_tolerations`` is ``None`` or has length 0.

    Returns
    -------
    client : object
        :py:class:`dask.distributed.Client` connected to cluster
    cluster : object
        Pre-configured :py:class:`dask_gateway.GatewayCluster`


    See Also
    --------
    :py:func:`get_micro_cluster` :
        A cluster with one-CPU workers
    :py:func:`get_standard_cluster` :
        The default cluster specification
    :py:func:`get_big_cluster` :
        A cluster with workers twice the size of the default
    :py:func:`get_giant_cluster` :
        A cluster with workers four times the size of the default
    """

    gateway = dask_gateway.Gateway()
    default_options = gateway.cluster_options()

    new_kwargs = kwargs.copy()

    if new_kwargs.get("cpus", 0) > 7.25:
        raise ValueError("Must specify ``cpus`` <= 7.25")

    # handle naming changes
    for k, v in kwargs.items():
        if k == "name":
            new_kwargs["worker_image"] = kwargs["name"]
            del new_kwargs["name"]
        elif k == "cred_path":
            if "cred_name" not in kwargs.keys():
                new_kwargs["cred_name"] = Path(v).stem
            del new_kwargs["cred_path"]
        elif k == "extra_pod_tolerations":
            if (
                "keep_default_tolerations" in kwargs.keys()
                and kwargs["keep_default_tolerations"] == False
            ):
                base_tols = {}
            else:
                base_tols = default_options.worker_tolerations
            new_kwargs.pop("keep_default_tolerations", None)
            new_kwargs["worker_tolerations"] = {
                **base_tols,
                **{
                    f"user_{key}": val
                    for key, val in enumerate(new_kwargs.pop("extra_pod_tolerations"))
                },
            }
        elif k not in list(default_options.keys()) + ["tag"]:
            raise KeyError(f"{k} not allowed as a kwarg when using dask-gateway")

    if "worker_image" in new_kwargs and "tag" in new_kwargs:
        raise ValueError("provide either `name` or `tag`, not both")

    if "tag" in new_kwargs:
        img, _ = default_options.worker_image.split(":")
        new_kwargs["worker_image"] = ":".join(img, new_kwargs["tag"])

    cluster = gateway.new_cluster(**new_kwargs)
    client = cluster.get_client()

    return client, cluster


def _get_cluster_dask_kubernetes(
    name=None,
    tag=None,
    extra_pip_packages=None,
    extra_conda_packages=None,
    memory_gb=None,
    nthreads=None,
    cpus=None,
    cred_name=None,
    cred_path=None,
    env_items=None,
    scaling_factor=1,
    dask_config_dict={},
    deploy_mode="local",
    idle_timeout=None,
    template_path="~/worker-template.yml",
    extra_worker_labels=None,
    extra_pod_tolerations=None,
    keep_default_tolerations=True,
    **kwargs,
):
    """

    **DEPRECATED (12/15/2020) **: Since we no longer maintain clusters using
    dask-kubernetes schedulers. Only dask-gateway is now supported.

    Start dask.kubernetes cluster and dask.distributed client

    All arguments are optional. If not provided, arguments will default to
    values provided in ``template_path``.

    Parameters
    ----------
    name : str, optional
        Name of worker image to use (e.g. ``rhodium/worker:latest``). If ``None``
        (default), default to worker specified in ``template_path``.
    tag : str, optional
        Tag of the worker image to use. Cannot be used in combination with
        ``name``, which should include a tag. If provided, overrides the
        tag of the image specified in ``template_path``. If ``None``
        (default), the full image specified in ``name`` or ``template_path``
        is used.
    extra_pip_packages : str, optional
        Extra pip packages to install on worker. Packages are installed
        using ``pip install extra_pip_packages``.
    extra_conda_packages :str, optional
        Extra conda packages to install on worker. Default channel is
        ``conda-forge``. Packages are installed using
        ``conda install -y -c conda-forge ${EXTRA_CONDA_PACKAGES}``.
    memory_gb : float, optional
        Memory to assign per 'group of workers', where a group consists of
        nthreads independent workers.
    nthreads : int, optional
        Number of independent threads per group of workers. Not sure if this
        should ever be set to something other than 1.
    cpus : float, optional
        Number of virtual CPUs to assign per 'group of workers'
    cred_name : str, optional
        Name of Google Cloud credentials file to use, equivalent to providing
        ``cred_path='/opt/gcsfuse_tokens/{}.json'.format(cred_name)``
    cred_path : str, optional
        Path to Google Cloud credentials file to use.
    env_items : dict, optional
        A dictionary of env variable 'name'-'value' pairs to append to the env
        variables included in ``template_path``, e.g.

        .. code-block:: python

            {
                'GOOGLE_APPLICATION_CREDENTIALS': '/opt/gcsfuse_tokens/rhg-data.json',
            }

    scaling_factor : float, optional
        scale the worker memory & CPU size using a constant multiplier of the
        specified worker. No constraints in terms of performance or cluster
        size are enforced - if you request too little the dask worker will not
        perform; if you request too much you may see an ``InsufficientMemory``
        or ``InsufficientCPU`` error on the google cloud Kubernetes console.
        Recommended scaling factors given our default ``~/worker-template.yml``
        specs are [0.5, 1, 2, 4].
    dask_config_dict : dict, optional
        Dask config parameters to modify from their defaults. A '.' is used
        to access progressive levels of the yaml structure. For instance, the
        dict could look like ``{'distributed.worker.profile.interval': '100ms'}``
    deploy_mode : str, optional
        Where to deploy the scheduler (on the same pod or a different pod)
    idle_timeout : str, optional
        Number of seconds without active communication with the client before the
        remote scheduler shuts down (ignored if ``deploy_mode=='local'``).
        Default is to not shut down for this reason.
    template_path : str, optional
        Path to worker template file. Default ``~/worker-template.yml``.
    extra_worker_labels : dict, optional
        Dictionary of kubernetes labels to apply to pods. None (default) results
        in no additional labels besides those in the template, as well as
        ``jupyter_user``, which is inferred from the ``JUPYTERHUB_USER``, or, if
        not set, the server's hostname.
    extra_pod_tolerations : list of dict, optional
        List of pod toleration dictionaries. For example, to match a node pool
        NoSchedule toleration, you might provide:

        .. code-block:: python

            extra_pod_tolerations=[
                {
                    "effect": "NoSchedule",
                    "key": "k8s.dask.org_dedicated",
                    "operator": "Equal",
                    "value": "worker-highcpu"
                },
                {
                    "effect": "NoSchedule",
                    "key": "k8s.dask.org/dedicated",
                    "operator": "Equal",
                    "value": "worker-highcpu"
                }
            ]

    keep_default_tolerations : bool, optional
        Whether to append (default) or replace the default tolerations. Ignored if
        ``extra_pod_tolerations`` is ``None`` or has length 0.

    Returns
    -------
    client : object
        :py:class:`dask.distributed.Client` connected to cluster
    cluster : object
        Pre-configured :py:class:`dask_kubernetes.KubeCluster`


    See Also
    --------
    :py:func:`get_micro_cluster` :
        A cluster with one-CPU workers
    :py:func:`get_standard_cluster` :
        The default cluster specification
    :py:func:`get_big_cluster` :
        A cluster with workers twice the size of the default
    :py:func:`get_giant_cluster` :
        A cluster with workers four times the size of the default

    """

    if (name is not None) and (tag is not None):
        raise ValueError("provide either `name` or `tag`, not both")

    # update dask settings
    dask.config.set(dask_config_dict)

    template_path = os.path.expanduser(template_path)

    with open(template_path, "r") as f:
        template = yml.load(f, Loader=yml.SafeLoader)

    # update labels with default and user-provided labels
    if ("metadata" not in template) or (template.get("metadata", {}) is None):
        template["metadata"] = {}

    if ("labels" not in template["metadata"]) or (
        template["metadata"]["labels"] is None
    ):
        template["metadata"]["labels"] = {}

    labels = template["metadata"]["labels"]

    if extra_worker_labels is not None:
        labels.update(extra_worker_labels)

    labels.update(
        {"jupyter_user": os.environ.get("JUPYTERHUB_USER", socket.gethostname())}
    )

    template["metadata"]["labels"] = labels

    if "tolerations" not in template["spec"]:
        template["spec"]["tolerations"] = []

    if (extra_pod_tolerations is not None) and (len(extra_pod_tolerations) > 0):
        if keep_default_tolerations:
            template["spec"]["tolerations"].extend(extra_pod_tolerations)
        else:
            template["spec"]["tolerations"] = extra_pod_tolerations

    container = template["spec"]["containers"][0]

    # replace the defualt image with the new one
    if name is not None:
        container["image"] = name
    if tag is not None:
        img, _ = container["image"].split(":")
        container["image"] = ":".join(img, tag)

    if extra_pip_packages is not None:
        container["env"].append(
            {"name": "EXTRA_PIP_PACKAGES", "value": extra_pip_packages}
        )

    if extra_conda_packages is not None:
        container["env"].append(
            {"name": "EXTRA_CONDA_PACKAGES", "value": extra_conda_packages}
        )

    if cred_path is not None:
        container["env"].append(
            {"name": "GOOGLE_APPLICATION_CREDENTIALS", "value": cred_path}
        )

    elif cred_name is not None:
        container["env"].append(
            {
                "name": "GOOGLE_APPLICATION_CREDENTIALS",
                "value": "/opt/gcsfuse_tokens/{}.json".format(cred_name),
            }
        )

    if env_items is not None:
        if isinstance(env_items, dict):
            [
                container["env"].append({"name": k, "value": v})
                for k, v in env_items.items()
            ]
        # allow deprecated passing of list of name/value pairs
        elif isinstance(env_items, Sequence):
            warnings.warn(
                "Passing of list of name/value pairs deprecated. "
                "Please pass a dictionary instead."
            )
            container["env"] = container["env"] + env_items
        else:
            raise ValueError("Expected `env_items` of type dict or sequence.")

    # adjust worker creation args
    args = container["args"]

    # set nthreads if provided
    nthreads_ix = args.index("--nthreads") + 1
    if nthreads is not None:
        args[nthreads_ix] = str(nthreads)
    nthreads = int(args[nthreads_ix])

    # then in resources
    resources = container["resources"]
    limits = resources["limits"]
    requests = resources["requests"]

    msg = "{} limits and requests do not match"

    if memory_gb is None:
        memory_gb = float(limits["memory"].strip("G"))
        mem_request = float(requests["memory"].strip("G"))
        assert memory_gb == mem_request, msg.format("memory")

    if cpus is None:
        cpus = float(limits["cpu"])
        cpu_request = float(requests["cpu"])
        assert cpus == cpu_request, msg.format("cpu")

    # now properly set the threads accessible by multi-threaded libraries
    # so that there's no competition between dask threads and the threads of
    # these libraries
    cpus_rounded = np.round(cpus)
    lib_threads = int(cpus_rounded / nthreads)
    for lib in ["OMP_NUM_THREADS", "MKL_NUM_THREADS", "OPENBLAS_NUM_THREADS"]:
        container["env"].append({"name": lib, "value": lib_threads})

    def format_request(x):
        return "{:04.2f}".format(np.floor(x * 100) / 100)

    # set memory-limit if provided
    mem_ix = args.index("--memory-limit") + 1
    args[mem_ix] = format_request(float(memory_gb) * scaling_factor) + "GB"

    limits["memory"] = format_request(float(memory_gb) * scaling_factor) + "G"

    requests["memory"] = format_request(float(memory_gb) * scaling_factor) + "G"

    limits["cpu"] = format_request(float(cpus) * scaling_factor)
    requests["cpu"] = format_request(float(cpus) * scaling_factor)

    # start cluster and client and return
    # need more time to connect to remote scheduler
    if deploy_mode == "remote":
        dask.config.set({"kubernetes.idle-timeout": idle_timeout})
    cluster = KubeCluster.from_dict(
        template, deploy_mode=deploy_mode, idle_timeout=None
    )

    client = dd.Client(cluster)

    return client, cluster


@_append_docstring()
def get_cluster(*args, **kwargs):
    if GATEWAY:
        if len(args) > 0:
            raise ValueError(
                "Positional args not allowed when using dask-gateway. If you are "
                "trying to pass worker name, use the ``name`` kwarg."
            )
        return _get_cluster_dask_gateway(**kwargs)
    return _get_cluster_dask_kubernetes(*args, **kwargs)


@_append_docstring()
def get_giant_cluster(*args, **kwargs):
    """
    Start a cluster with 4x the memory and CPU per worker relative to default
    """

    if GATEWAY:
        if len(args) > 0:
            raise ValueError(
                "Positional args not allowed when using dask-gateway. If you are "
                "trying to pass worker name, use the ``name`` kwarg."
            )
        return _get_cluster_dask_gateway(profile="giant", **kwargs)
    return _get_cluster_dask_kubernetes(*args, scaling_factor=4, **kwargs)


@_append_docstring()
def get_big_cluster(*args, **kwargs):
    """
    Start a cluster with 2x the memory and CPU per worker relative to default
    """
    if GATEWAY:
        if len(args) > 0:
            raise ValueError(
                "Positional args not allowed when using dask-gateway. If you are "
                "trying to pass worker name, use the ``name`` kwarg."
            )
        return _get_cluster_dask_gateway(profile="big", **kwargs)
    return _get_cluster_dask_kubernetes(*args, scaling_factor=2, **kwargs)


@_append_docstring()
def get_standard_cluster(*args, **kwargs):
    """
    Start a cluster with 1x the memory and CPU per worker relative to default
    """
    if GATEWAY:
        if len(args) > 0:
            raise ValueError(
                "Positional args not allowed when using dask-gateway. If you are "
                "trying to pass worker name, use the ``name`` kwarg."
            )
        return _get_cluster_dask_gateway(profile="standard", **kwargs)
    return _get_cluster_dask_kubernetes(*args, scaling_factor=1, **kwargs)


@_append_docstring()
def get_micro_cluster(*args, **kwargs):
    """
    Start a cluster with a single CPU per worker
    """
    if GATEWAY:
        if len(args) > 0:
            raise ValueError(
                "Positional args not allowed when using dask-gateway. If you are "
                "trying to pass worker name, use the ``name`` kwarg."
            )
        return _get_cluster_dask_gateway(profile="micro", **kwargs)
    return _get_cluster_dask_kubernetes(*args, scaling_factor=(0.97 / 1.75), **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `rhg_compute_tools` package."""

from time import sleep

import pytest
from dask_gateway import Gateway

from rhg_compute_tools import kubernetes


class LocalGateway(Gateway):
    def __init__(self):
        # initialize local gateway
        super().__init__("http://127.0.0.1:8000")


def monkeypatch_cluster(func):
    def inner(monkeypatch, mem, cpu, scale):
        def mock_KubeCluster_from_dict(dict_argument, **kwargs):
            return dict_argument

        def mock_dask_Client(cluster):
            return cluster

        monkeypatch.setattr("dask_gateway.Gateway", LocalGateway)

        monkeypatch.setattr("dask.distributed.Client", mock_dask_Client)

        return func(mem=mem, cpu=cpu, scale=scale)

    return inner


@pytest.mark.parametrize("mem,cpu,scale", [(None, None, None)])
@monkeypatch_cluster
def test_create_worker_gateway(mem=None, cpu=None, scale=None):

    # can only test options that are valid on a UnsafeLocalBackend (e.g. nothing that
    # has to do with kubernetes pod config)
    client, cluster = kubernetes._get_cluster_dask_gateway(
        profile="micro",
        cred_path="path/to/test_cred.json",
        env_items={"TEST_ITEM": "TEST_RESULT"},
    )

    cluster.scale(1)
    sleep(2)

    # check env_items worked
    def return_test_env_var(key):
        import os

        return os.environ[key]

    assert client.gather(
        client.map(return_test_env_var, ["GOOGLE_APPLICATION_CREDENTIALS", "TEST_ITEM"])
    ) == ["/opt/gcsfuse_tokens/test_cred.json", "TEST_RESULT"]

    # scale down gracefully
    cluster.scale(0), cluster.close(), client.close()


size_test_params = [(35, 7, None), (1, 6, None), (4, 1, None)]


scale_test_params = [
    (None, None, 0.5),
    (None, None, 1),
    (None, None, 2),
    (None, None, 4),
]



import matplotlib.cm
import numpy as np
from matplotlib.colors import LinearSegmentedColormap

try:
    _string_types = (str, unicode)
except NameError:
    _string_types = (str,)


def get_color_scheme(values, cmap=None, colors=None, levels=None, how=None):
    """
    Generate a norm and color scheme from data

    Parameters
    ----------
    values : array-like
        data to be plotted, from which to generate cmap and norm.
        This should be an array, DataArray, etc. that we can use
        to find the min/max and/or quantiles of the data.
    cmap : str, optional
        named matplotlib cmap (default inferred from data)
    colors : list-like, optional
        list of colors to use in a discrete colormap, or with which
        to create a custom color map
    levels : list-like, optional
        boundaries of discrete colormap, provide
    how : str, optional
        Optional setting form ``{'linear', 'log', 'symlog', None}``.
        Used to construct the returned ``norm`` object, which defines
        the way the colors map to values. By default, we the method is
        inferred from the ``values``.

    Returns
    -------
    cmap : object
        :py:class:`matplotlib.colors.cmap` color mapping
    norm : object
        :py:class:`matplotlib.colors.Normalize` instance using the provided
        values, levels, color specification, and "how" method

    """

    mini, maxi = float(values.min()), float(values.max())
    amax = max(abs(mini), abs(maxi))

    if (cmap is None) and (colors is not None):
        cmap = LinearSegmentedColormap.from_list("custom_cmap", colors)
    elif cmap is None:
        if (mini < 0) and (maxi > 0):
            cmap = matplotlib.cm.RdBu_r
        else:
            cmap = matplotlib.cm.viridis
    elif isinstance(cmap, _string_types):
        cmap = matplotlib.cm.get_cmap(cmap)

    if how is None and levels is not None:
        norm = matplotlib.colors.BoundaryNorm(levels, cmap.N)

    elif (how is None) or (how == "eq_hist"):
        if levels is None:
            levels = 11

        bin_edges = np.percentile(
            values[~np.isnan(values)], np.linspace(0, 100, levels)
        )

        norm = matplotlib.colors.BoundaryNorm(bin_edges, cmap.N)

    elif how == "log":
        norm = matplotlib.colors.LogNorm(vmin=mini, vmax=maxi)

    elif how == "symlog":
        norm = matplotlib.colors.SymLogNorm(
            vmin=-amax, vmax=amax, linthresh=(amax / 100)
        )

    elif how == "linear":
        norm = matplotlib.colors.Normalize(vmin=mini, vmax=maxi)

    else:
        raise ValueError(
            "color scheme `how` argument {} not recognized. "
            "choose from {eq_hist, log, symlog, linear} or "
            "provide `levels`".format(how)
        )

    return cmap, norm


def add_colorbar(ax, cmap="viridis", norm=None, orientation="vertical", **kwargs):
    """
    Add a colorbar to a plot, using a pre-defined cmap and norm

    Parameters
    ----------

    ax : object
        matplotlib axis object
    cmap : str or object, optional
        :py:class:`matplotlib.colors.cmap` instance or name of a
        registered cmap (default viridis)
    norm: object, optional
        :py:class:`matplotlib.colors.Normalize` instance. default is a linear
        norm between the min and max of the first plotted object.
    orientation : str, optional
        default 'vertical'
    **kwargs :
        passed to colorbar constructor
    """

    if norm is None:
        norm = matplotlib.colors.Normalize

    n_cmap = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)
    n_cmap.set_array([])
    cbar = ax.get_figure().colorbar(n_cmap, ax=ax, orientation=orientation, **kwargs)

    return cbar



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import re

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = re.sub(r"\(:issue:`[0-9]+`\)", "", history_file.read())

requirements = [
    "google-cloud-storage",
    "click",
    "dask-gateway",
    "pandas",
    "xarray",
    "matplotlib",
    "numpy",
    "bottleneck",
    # TODO: put package requirements here
]

setup(
    name="rhg_compute_tools",
    use_scm_version=True,
    description="Tools for using compute.rhg.com and compute.impactlab.org",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    author="Michael Delgado",
    author_email="mdelgado@rhg.com",
    url="https://github.com/RhodiumGroup/rhg_compute_tools",
    packages=find_packages(include=["rhg_compute_tools"]),
    include_package_data=True,
    package_data={
        "rhg_compute_tools": ["*.mplstyle"],
    },
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="rhg_compute_tools",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={
        "console_scripts": [
            "rctools = rhg_compute_tools.cli:rctools_cli",
        ]
    },
)
